import asyncio
from typing import Any, Dict, List, Optional
from theine import Memoize, Cache


@Memoize(1000, None)
def foo(id: int) -> Dict[str, int]:
    return {"id": id}


@foo.key
def _(id: int) -> str:
    return f"id-{id}"


class Bar:

    @Memoize(1000, None)
    def foo(self, id: int) -> Dict[str, int]:
        return {"id": id}

    @foo.key
    def _(self, id: int) -> str:
        return f"id-{id}"

    @Memoize(1000, None)
    async def async_foo(self, id: int) -> Dict[str, int]:
        await asyncio.sleep(1)
        return {"id": id}

    @async_foo.key
    def _(self, id: int) -> str:
        return f"id-{id}"


@Memoize(1000, None)
async def async_foo(id: int) -> Dict[str, int]:
    await asyncio.sleep(1)
    return {"id": id}


def run() -> None:
    v: Dict[str, int] = foo(12)
    bar = Bar()
    b: Dict[str, int] = bar.foo(13)

    client = Cache[int, int](1000)
    client.set(1, 1)
    v2, ok = client.get(1)
    vt: Optional[int] = v2
    okk: bool = ok


async def run_async() -> None:
    v: Dict[str, int] = await async_foo(12)
    bar = Bar()
    b: Dict[str, int] = await bar.async_foo(13)



SECRET_KEY = "django_tests_secret_key"
CACHES = {
    "default": {
        "BACKEND": "theine.adapters.django.Cache",
        "TIMEOUT": 60,
        "OPTIONS": {"MAX_ENTRIES": 1000, "POLICY": "tlfu"},
    },
}

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
]

USE_TZ = False



from typing import Any, Dict, List, Optional
from theine import Memoize, Cache


@Memoize(1000, None)
def foo(id: int) -> Dict[str, int]:
    return {"id": id}


@foo.key
def _(id: int, name: str) -> str:
    return f"id-{id}"


class Bar:

    @Memoize(1000, None)
    def foo(self, id: int) -> Dict[str, int]:
        return {"id": id}

    @foo.key
    def _(self, id: int, name: str) -> str:
        return f"id-{id}"


def run() -> None:
    v: Dict[str, int] = foo(12, 13)
    bar = Bar()
    b: Dict[str, int] = bar.foo(12, 13)

    client = Cache[int, int](1000)
    client.set("a", 1)
    client.set(1, "a")
    v2, ok = client.get(1)
    vt: Optional[str] = v2
    okk: str = ok



import sys
import sysconfig
import threading
from datetime import timedelta
from random import randint
from time import sleep

import pytest
from bounded_zipf import Zipf  # type: ignore[import]

from theine.theine import Cache

is_freethreaded = bool(sysconfig.get_config_var("Py_GIL_DISABLED"))


@pytest.fixture(params=[True, False])
def nolock(request):
    return request.param

def test_set(nolock) -> None:
    cache = Cache(100, nolock)
    for i in range(20):
        key = f"key:{i}"
        cache.set(key, key)
    for i in range(20):
        key = f"key:{i}"
        v, ok = cache.get(key)
        assert ok
        assert v == key
    for i in range(20):
        key = f"key:{i}"
        cache.set(key, key)
    for i in range(20):
        key = f"key:{i}"
        v, ok = cache.get(key)
        assert ok
        assert v == key
    for i in range(100):
        key = f"key:{i}:v2"
        cache.set(key, key)

    cache._force_drain_write()
    assert len(cache) == 100
    sleep(1)
    cache.set("foo", "bar")
    cache._force_drain_write()
    assert len(cache) == 100


def test_set_cache_size(nolock) -> None:
    cache = Cache(500, nolock)
    for _ in range(100000):
        i = randint(0, 100000)
        cache.set(f"key:{i}", i)
    cache._force_drain_write()
    assert len(cache) == 500


def test_set_with_ttl(nolock) -> None:
    cache = Cache(500, nolock)
    for i in range(30):
        key = f"key:{i}"
        cache.set(key, key, timedelta(seconds=i + 1))
        key = f"key:{i}:2"
        cache.set(key, key, timedelta(seconds=i + 100))
    cache._force_drain_write()
    assert len(cache) == 60
    current = 60
    counter = 0

    while True:
        sleep(5)
        counter += 1
        assert len(cache) < current
        current = len(cache)
        if current <= 30:
            break
    assert counter < 10

    for i in range(30):
        v, ok = cache.get(f"key:{i}:2")
        assert ok


def test_set_with_ttl_multi_instances(nolock) -> None:
    caches = []
    for i in range(30):
        caches.append(Cache(500, nolock))

    for cache in caches:
        for i in range(30):
            key = f"key:{i}"
            cache.set(key, key, timedelta(seconds=1 + randint(0, 5)))
        cache._force_drain_write()
        assert len(cache) == 30

    sleep(8)
    for cache in caches:
        assert len(cache) == 0


class SameHash:

    def __init__(self, i):
        self.i = i

    def __hash__(self):
        return 3245671

    def __eq__(self, other):
        return other == self.i


def test_collision(nolock) -> None:
    cache = Cache(500, nolock)
    for i in range(30):
        e = SameHash(i)
        cache.set(e, e, timedelta(seconds=i + 5))
    # only the last entry exists because all entries have same hash
    cache._force_drain_write()
    assert len(cache) == 1
    assert (cache.core.len()) == 1
    obj, ok = cache.get(SameHash(29))
    assert ok
    assert obj.i == 29


def test_delete(nolock) -> None:
    cache = Cache(100, nolock)
    for i in range(20):
        key = f"key:{i}"
        cache.set(key, key)
    cache.delete("key:1")
    cache.delete("key:3")
    cache.delete("key:5")
    cache._force_drain_write()
    assert len(cache) == 17


class Foo:
    def __init__(self, id: int):
        self.id = id


def test_hashable_key(nolock) -> None:
    cache = Cache(100, nolock)
    foos = [Foo(i) for i in range(20)]
    for foo in foos:
        cache.set(foo, foo)
    for foo in foos:
        cached, ok = cache.get(foo)
        assert cached is foo
    cache.delete(foos[3])
    cached, ok = cache.get(foos[3])
    assert not ok


def test_set_with_ttl_hashable(nolock) -> None:
    cache = Cache(500, nolock)
    foos = [Foo(i) for i in range(30)]
    for i in range(30):
        cache.set(foos[i], foos[i], timedelta(seconds=i + 1))
    cache._force_drain_write()
    assert len(cache) == 30
    current = 30
    counter = 0
    while True:
        sleep(5)
        counter += 1
        assert len(cache) < current
        current = len(cache)
        if current == 0:
            break


def test_expire_proactively(nolock) -> None:
    cache = Cache(50000, nolock)
    for i in range(50000):
        cache.set((i, 2), i, timedelta(seconds=randint(5, 10)))
    current = 50000
    cache._force_drain_write()
    while True:
        sleep(1)
        assert len(cache) <= current
        current = len(cache)
        if current == 0:
            break
    assert len(cache.core.keys()) == 0


def test_clear_cache(nolock) -> None:
    cache = Cache(500, nolock)
    for i in range(5000):
        cache.set((i, 2), i, timedelta(seconds=randint(5, 10)))
    sleep(1)

    cache.clear()
    assert len(cache) == 0
    assert cache.core.len() == 0


def test_close_cache(nolock) -> None:
    for _ in range(10):
        cache = Cache(500, nolock)
        cache.set("foo", "bar", timedelta(seconds=60))
        cache.close()
        sleep(3)
        assert cache._maintainer.done() is True


def test_cache_stats(nolock) -> None:
    cache = Cache(5000, nolock)
    assert cache.max_size == 5000
    assert len(cache) == 0
    z = Zipf(1.0001, 10, 20000)
    for _ in range(20000):
        i = z.get()
        key = f"key:{i}"
        v, ok = cache.get(key)
        if not ok:
            cache.set(key, key)
    cache._force_drain_write()
    stats = cache.stats()
    assert stats.hit_count > 0
    assert stats.miss_count > 0
    assert stats.request_count == stats.hit_count + stats.miss_count
    assert stats.hit_rate > 0.5
    assert stats.hit_rate < 1
    assert stats.hit_rate == stats.hit_count / stats.request_count


def zipf_key_gen(total):
    z = Zipf(1.01, 9.0, 50000 * 1000)
    for _ in range(total):
        yield z.get()


def test_zipf(nolock) -> None:
    miss = 0
    cache = Cache(50000, nolock)
    for key in zipf_key_gen(2000000):
        v = cache.get(key)
        if v[1]:
            assert key == v[0]
        else:
            miss += 1
            cache.set(key, key)
    stats = cache.stats()
    assert stats.request_count == 2000000
    assert stats.hit_rate > 0.5 and stats.hit_rate < 0.6
    assert 1 - (miss / 2000000) > 0.5 and 1 - (miss / 2000000) < 0.6


def test_zipf_correctness(nolock) -> None:
    for size in [500, 2000, 10000, 50000]:
        cache = Cache(size, nolock)
        for key in zipf_key_gen(1000000):
            v = cache.get(key)
            if v[1]:
                assert key == v[0]
            else:
                cache.set(key, key)
        cache._force_drain_write()
        total = 0
        for shard in cache._shards:
            total += len(shard._map)
        info = cache.core.debug_info()
        assert total == info.len
        assert info.window_len + info.probation_len + info.protected_len == total
        keys = cache.core.keys()
        assert len(keys) == total
        for kh in cache.core.keys():
            idx = kh & (cache._shard_count - 1)
            assert kh in cache._shards[idx]._key_map


def test_zipf_correctness_parallel() -> None:
    def run(client):
        if is_freethreaded:
            assert sys._is_gil_enabled() is False

        keys = list(zipf_key_gen(500000))
        for key in keys:
            v = client.get(key)
            if v[1]:
                assert key == v[0]
            else:
                client.set(key, key)

    for size in [500, 2000, 10000, 50000]:
        cache = Cache(size)
        threads = []
        for start in range(4):
            thread = threading.Thread(target=run, args=[cache])
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        cache._force_drain_write()
        total = 0
        for shard in cache._shards:
            total += len(shard._map)
        info = cache.core.debug_info()
        assert total == info.len
        assert info.window_len + info.probation_len + info.protected_len == total
        keys = cache.core.keys()
        assert len(keys) == total
        for kh in cache.core.keys():
            idx = kh & (cache._shard_count - 1)
            assert kh in cache._shards[idx]._key_map



from datetime import timedelta
from threading import Lock
from typing import Any, Callable, Dict, Optional, Union, cast

from django.core.cache.backends.base import BaseCache, DEFAULT_TIMEOUT

from theine import Cache as Theine
from theine.theine import sentinel

KEY_TYPE = Union[str, Callable[..., str]]
VALUE_TYPE = Any
VERSION_TYPE = Optional[int]


class Cache(BaseCache):
    def __init__(self, name: str, params: Dict[str, Any]):
        super().__init__(params)
        self.cache = Theine[Any, Any](self._max_entries)

    def _timeout_seconds(
        self, timeout: "Optional[Union[float, DEFAULT_TIMEOUT]]"
    ) -> float:
        if timeout == DEFAULT_TIMEOUT:
            return cast(float, self.default_timeout)
        return cast(float, timeout)

    def add(
        self,
        key: KEY_TYPE,
        value: VALUE_TYPE,
        timeout: Optional[float] = DEFAULT_TIMEOUT,
        version: VERSION_TYPE = None,
    ) -> bool:
        data = self.get(key, sentinel, version)
        if data is not sentinel:
            return False
        key = self.make_key(key, version)
        self.cache.set(
            key,
            value,
            (
                timedelta(seconds=cast(float, timeout))
                if timeout is not DEFAULT_TIMEOUT
                else None
            ),
        )
        return True

    def get(
        self,
        key: KEY_TYPE,
        default: Optional[VALUE_TYPE] = None,
        version: VERSION_TYPE = None,
    ) -> Optional[VALUE_TYPE]:
        key = self.make_key(key, version)
        v, ok = self.cache.get(key)
        if not ok:
            return default
        return v

    def set(
        self,
        key: KEY_TYPE,
        value: VALUE_TYPE,
        timeout: Optional[float] = DEFAULT_TIMEOUT,
        version: VERSION_TYPE = None,
    ) -> None:
        to = self._timeout_seconds(timeout)
        if to is not None and to <= 0:
            self.delete(key)
            return
        key = self.make_key(key, version)
        self.cache.set(
            key,
            value,
            timedelta(seconds=to) if to is not None else None,
        )

    def touch(
        self,
        key: KEY_TYPE,
        timeout: Optional[float] = DEFAULT_TIMEOUT,
        version: VERSION_TYPE = None,
    ) -> bool:
        data = self.get(key, sentinel, version)
        if data is sentinel:
            return False
        nkey = self.make_key(key, version)
        if timeout is not DEFAULT_TIMEOUT and timeout is not None and timeout <= 0:
            self.cache.delete(nkey)
            return True
        to = self._timeout_seconds(timeout)
        with Lock():
            self.cache._access(nkey, timedelta(seconds=to) if to is not None else None)
        return True

    def delete(self, key: KEY_TYPE, version: VERSION_TYPE = None) -> bool:
        key = self.make_key(key, version)
        return self.cache.delete(key)

    def clear(self) -> None:
        self.cache.clear()



import concurrent.futures
import gzip
import struct
from datetime import timedelta
from random import randint
from threading import Thread
from typing import Callable, Iterator, List

import matplotlib.pyplot as plt
import numpy as np
import zstandard as zstd
from bounded_zipf import Zipf
from cachetools import FIFOCache, LRUCache
from scipy.interpolate import PchipInterpolator

from theine import Cache

plt.style.use("ggplot")


GET = "GET"
SET = "SET"


def zipf_key_gen(name: str) -> Iterator:
    z = Zipf(1.001, 10, 50000000)
    for _ in range(1000000):
        yield f"{z.get()}", GET


def arc_key_gen(name: str) -> Iterator:
    with gzip.open(f"benchmarks/trace/{name.lower()}.gz", "rt") as f:
        for line in f:
            tmp = line.split(" ")
            base = int(tmp[0])
            count = int(tmp[1])
            for i in range(count):
                yield str(base + i), GET


def oracle_general_gen(name: str) -> Iterator:
    path = ""
    if name == "WIKI CDN":
        path = "benchmarks/trace/wiki_2019t.oracleGeneral.zst"
    if name == "Twitter Cluster 52":
        path = "benchmarks/trace/cluster52.oracleGeneral.sample10.zst"
    with open(path, "rb") as f:
        dctx = zstd.ZstdDecompressor()
        with dctx.stream_reader(f) as reader:
            buffer_size = 24
            buffer = bytearray(buffer_size)

            while True:
                try:
                    num_read = reader.readinto(buffer)
                    if num_read < buffer_size:
                        return
                except EOFError:
                    return
                except Exception as err:
                    raise Exception(f"Wrapped error: {err}")

                # Interpret bytes 4 through the end as a little-endian unsigned 64-bit integer
                id = struct.unpack_from("<Q", buffer, offset=4)[0]
                yield str(id), GET


def lirs_key_gen(name: str) -> Iterator:
    with gzip.open("benchmarks/trace/loop.gz", "rt") as f:
        for line in f:
            k = line.strip()
            yield k, GET


def corda_key_gen(name: str) -> Iterator:
    buffer_size = 8
    buffer = bytearray(buffer_size)
    with gzip.open("benchmarks/trace/trace_vaultservice_large.gz", "rb") as reader:
        while True:
            try:
                num_read = reader.readinto(buffer)
                if num_read < buffer_size:
                    return
            except EOFError:
                return
            except Exception as err:
                raise Exception(f"Wrapped error: {err}")

            id = struct.unpack_from(">Q", buffer)[0]
            yield str(id), GET


def bench_theine(cap: int, gens: List[Callable[..., Iterator]], name: str):
    counter = 0
    miss = 0

    cache = Cache(cap, True)

    for gen in gens:
        for key, op in gen(name):
            counter += 1
            if op == GET:
                v = cache.get(key)
                if v[1]:
                    assert key == v[0]
                else:
                    miss += 1
                    cache.set(key, key)
            else:
                cache.set(key, key)

            if counter % 100000 == 0:
                print(".", end="", flush=True)
    print("")
    hr = (counter - miss) / counter
    print(f"---- theine hit ratio: {hr:.3f}")
    return hr


def bench_cachetools(
    policy: str, cap: int, gens: List[Callable[..., Iterator]], name: str
):
    counter = 0
    miss = 0

    if policy == "LRU":
        cache = LRUCache(cap)
    elif policy == "FIFO":
        cache = FIFOCache(cap)

    for gen in gens:
        for key, op in gen(name):
            counter += 1
            if op == GET:
                v = cache.get(key)
                if v is not None:
                    assert key == v
                else:
                    miss += 1
                    cache[key] = key
            else:
                cache[key] = key

            if counter % 100000 == 0:
                print(".", end="", flush=True)
    print("")
    hr = (counter - miss) / counter
    print(f"---- cachetools {policy} hit ratio: {hr:.3f}")
    return hr


def _infinit_run(cache: Cache[int, int]):
    z = Zipf(1.001, 10, 100000000)
    count = 0
    while True:
        key = z.get()
        data, exists = cache.get(key)
        if not exists:
            cache.set(key, key, timedelta(seconds=randint(30, 20000)))
        else:
            assert data == key
        count += 1
        if count % 100000 == 0:
            print(".", end="", flush=True)


def infinit_run_parallel(cap: int):
    cache = Cache(500000)

    threads = []
    for _ in range(6):
        thread = Thread(target=_infinit_run, args=[cache])
        thread.start()
        threads.append(thread)
    for t in threads:
        t.join()


def init_plot(title):
    fig, ax = plt.subplots()
    ax.set_xlabel("capacity")
    ax.set_ylabel("hit ratio")
    ax.set_title(title)
    fig.set_figwidth(12)
    return fig, ax


def bench_and_plot(caps, key_gens, name):
    x = []
    y_tlfu = []
    y_cachetools_lru = []
    y_cachetools_fifo = []

    for cap in caps:
        print(f"======= {name} cache size: {cap} =======")
        x.append(cap)
        y_tlfu.append(bench_theine(cap, key_gens, name))
        y_cachetools_lru.append(bench_cachetools("LRU", cap, key_gens, name))
        y_cachetools_fifo.append(bench_cachetools("FIFO", cap, key_gens, name))

    fig, ax = init_plot(f"Hit Ratios - {name}")
    # Smooth the lines using np.interp for interpolation
    x_new = np.linspace(min(x), max(x), 300)

    tlfu_spl = PchipInterpolator(x, y_tlfu)
    clru_spl = PchipInterpolator(x, y_cachetools_lru)
    cfifo_spl = PchipInterpolator(x, y_cachetools_fifo)
    y_tlfu_smooth = tlfu_spl(x_new)
    y_clru_smooth = clru_spl(x_new)
    y_cfifo_smooth = cfifo_spl(x_new)

    ax.plot(x_new, y_tlfu_smooth, "b-", label="theine")
    ax.plot(x_new, y_clru_smooth, "r-", label="cachetools-lru")
    ax.plot(x_new, y_cfifo_smooth, "c-", label="cachetools-fifo")
    ax.plot(x, y_tlfu, "bo")
    ax.plot(x, y_cachetools_lru, "ro")
    ax.plot(x, y_cachetools_fifo, "co")

    ax.legend()
    fig.savefig(f"benchmarks/{name.lower()}.png", dpi=200)


def bench_and_plot_parallel(caps, key_gens, name):
    x = []
    y_tlfu = []
    y_cachetools_lru = []
    y_cachetools_fifo = []

    futures = {}
    with concurrent.futures.ThreadPoolExecutor(max_workers=12) as executor:
        for cap in caps:
            futures[executor.submit(bench_theine, cap, key_gens, name)] = (
                "theine",
                cap,
            )
            futures[executor.submit(bench_cachetools, "LRU", cap, key_gens, name)] = (
                "lru",
                cap,
            )
            futures[executor.submit(bench_cachetools, "FIFO", cap, key_gens, name)] = (
                "fifo",
                cap,
            )

        m = {}
        for future in concurrent.futures.as_completed(futures):
            tp, cap = futures[future]
            data = future.result()
            if cap not in m:
                m[cap] = [0, 0, 0]
            if tp == "theine":
                m[cap][0] = data
            elif tp == "lru":
                m[cap][1] = data
            elif tp == "fifo":
                m[cap][2] = data

    keys = list(m.keys())
    keys.sort()
    for key in keys:
        x.append(key)
        v = m[key]
        y_tlfu.append(v[0])
        y_cachetools_lru.append(v[1])
        y_cachetools_fifo.append(v[2])

    fig, ax = init_plot(f"Hit Ratios - {name}")
    # Smooth the lines using np.interp for interpolation
    x_new = np.linspace(min(x), max(x), 300)

    tlfu_spl = PchipInterpolator(x, y_tlfu)
    clru_spl = PchipInterpolator(x, y_cachetools_lru)
    cfifo_spl = PchipInterpolator(x, y_cachetools_fifo)
    y_tlfu_smooth = tlfu_spl(x_new)
    y_clru_smooth = clru_spl(x_new)
    y_cfifo_smooth = cfifo_spl(x_new)

    ax.plot(x_new, y_tlfu_smooth, "b-", label="theine")
    ax.plot(x_new, y_clru_smooth, "r-", label="cachetools-lru")
    ax.plot(x_new, y_cfifo_smooth, "c-", label="cachetools-fifo")
    ax.plot(x, y_tlfu, "bo")
    ax.plot(x, y_cachetools_lru, "ro")
    ax.plot(x, y_cachetools_fifo, "co")

    ax.legend()
    fig.savefig(f"benchmarks/{name.lower()}.png", dpi=200)


# infinit_run_parallel(50000)

bench_and_plot(
    [500, 1000, 2000, 5000, 10_000, 20_000, 40_000, 80_000], [zipf_key_gen], "Zipf"
)
bench_and_plot(
    [
        1_000_000,
        2_000_000,
        3_000_000,
        4_000_000,
        5_000_000,
        6_000_000,
        7_000_000,
        8_000_000,
    ],
    [arc_key_gen],
    "DS1",
)
bench_and_plot(
    [100_000, 200_000, 300_000, 400_000, 500_000, 600_000, 700_000, 800_000],
    [arc_key_gen],
    "S3",
)

bench_and_plot(
    [25_000, 50_000, 100_000, 200_000, 300_000, 400_000, 500_000, 600_000],
    [arc_key_gen],
    "P3",
)

bench_and_plot(
    [10_000, 20_000, 30_000, 40_000, 50_000, 60_000, 70_000, 80_000],
    [arc_key_gen],
    "P8",
)

bench_and_plot(
    [250, 500, 750, 1000, 1250, 1500, 1750, 2000],
    [arc_key_gen],
    "OLTP",
)

bench_and_plot(
    [250, 500, 750, 1000, 1250, 1500, 1750, 2000],
    [lirs_key_gen],
    "LIRS",
)

bench_and_plot(
    [256, 512, 1024, 2048],
    [corda_key_gen],
    "VAULT",
)

bench_and_plot(
    [256, 512, 1024, 2048],
    [
        corda_key_gen,
        lirs_key_gen,
        lirs_key_gen,
        lirs_key_gen,
        lirs_key_gen,
        lirs_key_gen,
        corda_key_gen,
    ],
    "MIX",
)

bench_and_plot(
    [500, 1000, 1500, 2000],
    [oracle_general_gen],
    "Twitter Cluster 52",
)



from threading import Lock
from typing import List, Tuple, Callable


BufferSize = 16


class WriteBuffer:
    def __init__(
        self,
        send_to_core: Callable[[List[Tuple[int, int]]], List[int]],
        remove_keys: Callable[[List[int]], None],
        core_mutex: Lock,
        nolock: bool = False,
    ):
        # tuple: (key_hash, ttl)
        self.buffer: List[Tuple[int, int]] = [(0, 0) for _ in range(BufferSize)]
        self.mutex = Lock()
        self.core_mutex = core_mutex
        self.send_to_core = send_to_core
        self.remove_keys = remove_keys
        self.tail = 0
        self.nolock = nolock

    def add(self, key_hash: int, ttl: int) -> None:
        self.nolock or self.mutex.acquire()
        self.buffer[self.tail] = (key_hash, ttl)
        self.tail += 1
        # if tail < bufferSize, buffer is not full yet, we try to schedule a clear if core is not busy
        # if tail >= BufferSize: buffer is full and schedule a clear immediately
        if self.nolock or self.core_mutex.acquire(self.tail >= BufferSize):
            wb = self.buffer[: self.tail]
            self.tail = 0
            if not self.nolock:
                self.mutex.release()
            evicted = self.send_to_core(wb)
            if not self.nolock:
                self.core_mutex.release()
            self.remove_keys(evicted)
        else:
            if not self.nolock:
                self.mutex.release()



import asyncio
import inspect
import itertools
import os
import sysconfig
import time
from dataclasses import dataclass
from datetime import timedelta
from functools import _make_key, update_wrapper
from threading import Event, Lock, Thread
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Hashable,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
    no_type_check,
    overload,
)

from theine_core import TlfuCore, spread
from typing_extensions import Concatenate, ParamSpec, Protocol

from theine.models import KT, VT, CacheStats, Entry
from theine.striped_buffer import StripedBuffer
from theine.utils import round_up_power_of_2
from theine.write_buffer import WriteBuffer

S = TypeVar("S", contravariant=True)
P = ParamSpec("P")

sentinel = object()

_maintainer_loop = asyncio.new_event_loop()


def _maintance() -> None:
    asyncio.set_event_loop(_maintainer_loop)
    _maintainer_loop.run_forever()


_maintainer_thread = Thread(target=_maintance, daemon=True)
_maintainer_thread.start()


@dataclass
class EventData:
    event: Event
    data: Any


# https://github.com/python/cpython/issues/90780
# use event to protect from thundering herd
class CachedAwaitable:
    def __init__(
        self, awaitable: Awaitable[Any], on_error: Callable[[BaseException], None]
    ) -> None:
        self.awaitable = awaitable
        self.future: Optional[Awaitable[Any]] = None
        self.result = sentinel
        self.on_error = on_error
        self.exception: Optional[BaseException] = None

    def __await__(self) -> Any:
        if self.result is not sentinel:
            return self.result
        elif self.exception:
            raise self.exception

        if self.future is None:
            try:
                self.future = asyncio.Future()
                result = yield from self.awaitable.__await__()
                self.result = result
                self.future.set_result(self.result)
                self.future = None
                return result
            except BaseException as e:
                self.exception = e
                if self.future:
                    self.future.set_exception(e)
                self.future = None
                self.on_error(e)
                raise e
        else:
            yield from self.future.__await__()
        return self.result


class Key:
    def __init__(self) -> None:
        self.key: Optional[str] = None
        self.event = Event()


class Cached(Protocol[S, P, VT]):
    _cache: "Cache[Hashable, VT]"

    @overload
    def key(self, fn: Callable[P, Hashable]) -> None: ...

    @overload
    def key(self, fn: Callable[Concatenate[S, P], Hashable]) -> None: ...

    @overload
    def __call__(self, _arg_first: S, *args: P.args, **kwargs: P.kwargs) -> VT: ...

    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> VT: ...

    def cache_stats(self) -> CacheStats: ...


@no_type_check
def Wrapper(
    fn: Callable,
    ttl: Optional[timedelta],
    cache: "Cache",
    typed: bool,
    is_async: bool,
):
    _key_func = None
    _events = {}
    _func = fn
    _cache = cache
    _ttl = ttl
    _typed = typed
    _auto_key = True
    _is_async = is_async

    def key(fn) -> None:
        nonlocal _key_func
        nonlocal _auto_key
        _key_func = fn
        _auto_key = False

    def cache_stats() -> CacheStats:
        return _cache.stats()

    def fetch(*args, **kwargs):
        if _auto_key:
            key = _make_key(args, kwargs, _typed)
        else:
            key = _key_func(*args, **kwargs)

        if _is_async:
            result, ok = _cache.get(key)
            if not ok:
                result = CachedAwaitable(
                    _func(*args, **kwargs), lambda _: _cache.delete(key)
                )
                _cache.set(key, result, _ttl)
            return result

        return _cache._get_or_compute(key, lambda: fn(*args, **kwargs))

    fetch._cache = _cache
    _cache.with_loader(fn, ttl)
    fetch.key = key
    fetch.cache_stats = cache_stats
    return fetch


class Memoize:
    """
    Memoize decorator to cache function results. This decorator has 2 modes, first one custom-key mode, this
    is also the recommended mode. You must specify the key function manually. Second one is auto-key mode,
    Theine will generate a key for you based on your function inputs.

    :param size: cache size.
    :param ttl: timedelta to store the function result. Default is None which means no expiration.
    :param typed: Only valid with auto-key mode. If typed is set to true,
        function arguments of different types will be cached separately.
        If typed is false, the implementation will usually regard them as equivalent calls and only cache
        a single result. (Some types such as str and int may be cached separately even when typed is false.)
    :param nolock: disables thread locking for get and set operations. Defaults to False. Enable this only if your code does not use multi-threading.
    """

    def __init__(
        self,
        size: int,
        ttl: Optional[timedelta],
        typed: bool = False,
        nolock: bool = False,
    ):
        self.cache = Cache[Hashable, Any](size, nolock)
        self.ttl = ttl
        self.typed = typed

    def __call__(self, fn: Callable[Concatenate[S, P], VT]) -> Cached[S, P, VT]:
        wrapper = Wrapper(
            fn, self.ttl, self.cache, self.typed, inspect.iscoroutinefunction(fn)
        )
        return cast(Cached[S, P, VT], update_wrapper(wrapper, fn))


class Shard(Generic[KT, VT]):

    def __init__(self, nolock: bool) -> None:
        self._map: Dict[KT, Entry[VT]] = {}
        # key map is used to find evicted entries in _map because policy returns hashed key value
        self._key_map: Dict[int, KT] = {}
        self._mutex = Lock()
        self._hits = itertools.count()
        self._misses = itertools.count()
        self._nolock = nolock

    def get(self, key: KT, key_hash: int) -> Tuple[Optional[VT], bool]:
        try:
            # https://docs.python.org/3/howto/free-threading-python.html
            # It’s recommended to use the threading.Lock or other synchronization primitives instead of relying on the internal locks of built-in types, when possible.
            self._nolock or self._mutex.acquire()
            entry = self._map[key]
            if entry.expire > 0 and entry.expire <= time.monotonic_ns():
                next(self._misses)
                self._map.pop(key, sentinel)
                self._key_map.pop(key_hash, sentinel)
                return (None, False)
            else:
                next(self._hits)
                return (entry.value, True)
        except KeyError:
            next(self._misses)
            return (None, False)
        except Exception as err:
            raise err
        finally:
            if not self._nolock:
                self._mutex.release()

    # used only in the cache decorator. self._hits and self._misses are not updated here, as those values are recorded when entering the decorator
    def _get_nolock(self, key: KT, key_hash: int) -> Tuple[Optional[VT], bool]:
        try:
            entry = self._map[key]
            if entry.expire > 0 and entry.expire <= time.monotonic_ns():
                self._map.pop(key, sentinel)
                self._key_map.pop(key_hash, sentinel)
                return (None, False)
            else:
                return (entry.value, True)
        except KeyError:
            return (None, False)

    def _set_nolock(self, key: KT, key_hash: int, value: VT, ttl: int) -> bool:
        # remove exist first if key hash collision
        # not policy update because same hash means same key in policy
        try:
            removed = self._key_map.pop(key_hash)
        except KeyError:
            pass
        else:
            self._map.pop(removed)

        expire = 0
        if ttl > 0:
            expire = time.monotonic_ns() + ttl
        self._map[key] = Entry(value, expire)
        self._key_map[key_hash] = key
        return True

    def set(self, key: KT, key_hash: int, value: VT, ttl: int) -> bool:
        try:
            self._nolock or self._mutex.acquire()
            # remove exist first if key hash collision
            # not policy update because same hash means same key in policy
            try:
                removed = self._key_map.pop(key_hash)
            except KeyError:
                pass
            else:
                self._map.pop(removed)

            expire = 0
            if ttl > 0:
                expire = time.monotonic_ns() + ttl
            self._map[key] = Entry(value, expire)
            self._key_map[key_hash] = key
            return True
        except Exception as e:
            raise e
        finally:
            if not self._nolock:
                self._mutex.release()

    def set_ttl(self, key: KT, ttl: int) -> None:
        try:
            self._nolock or self._mutex.acquire()
            try:
                entry = self._map[key]
            except KeyError:
                return
            else:
                expire = 0
                if ttl > 0:
                    expire = time.monotonic_ns() + ttl
                entry.expire = expire
        except Exception as e:
            raise e
        finally:
            if not self._nolock:
                self._mutex.release()

    def remove(self, key_hash: int) -> bool:
        try:
            self._nolock or self._mutex.acquire()
            try:
                key = self._key_map.pop(key_hash)
            except KeyError:
                return False
            else:
                try:
                    del self._map[key]
                except KeyError:
                    return False
                else:
                    return True
        except Exception as e:
            raise e
        finally:
            if not self._nolock:
                self._mutex.release()

    # check expire before remove, used in timer wheel expire only,
    # set(key, ttl1) -> ttl1_sync_to_policy -> set(key, ttl2) -> ttl1 expired -> key_removed -> ttl2_sync_to_policy
    # the key is already removed when policy aware ttl changed, make ttl2 no effect
    def remove_expired(self, key_hash: int) -> None:
        try:
            self._nolock or self._mutex.acquire()
            try:
                key = self._key_map[key_hash]
            except KeyError:
                pass
            else:
                try:
                    self._map[key]
                except KeyError:
                    pass
                else:
                    # expired, removed
                    self._key_map.pop(key_hash, sentinel)
                    self._map.pop(key, sentinel)
        except Exception as e:
            raise e
        finally:
            if not self._nolock:
                self._mutex.release()

    def clear(self) -> None:
        with self._mutex:
            self._map.clear()
            self._key_map.clear()

    def __len__(self) -> int:
        with self._mutex:
            return len(self._map)


class Cache(Generic[KT, VT]):
    """
    Create new Theine cache store and use API with this class. This class is not thread-safe.

    :param policy: eviction policy, "tlfu", "lru" and "clockpro" are the only supported now.
    :param size: cache size.
    """

    _loader: Optional[Callable[[KT], VT]]

    _ttl: Optional[timedelta]

    def __init__(self, size: int, nolock: bool = False):
        if bool(sysconfig.get_config_var("Py_GIL_DISABLED")):
            nolock = False
        shard_count = round_up_power_of_2(os.cpu_count() or 4)

        if shard_count < 16:
            shard_count = 16
        elif shard_count > 128:
            shard_count = 128

        self._shards: Tuple[Shard[KT, VT], ...] = tuple(
            [Shard(nolock) for _ in range(shard_count)]
        )
        self.core = TlfuCore(size)
        self._closed = False
        self._total = 0
        self._hit = 0
        self.max_size = size
        self._shard_count = shard_count
        self.max_int64 = (1 << 64) - 1
        # core is single thread, all core operation must hold this mutex
        self._core_mutex = Lock()
        self._read_buffer = StripedBuffer(self._drain_read, nolock)
        self._write_buffer = WriteBuffer(
            self._drain_write, self._remove_keys, self._core_mutex, nolock
        )
        self._maintainer = asyncio.run_coroutine_threadsafe(
            self.maintenance(), loop=_maintainer_loop
        )
        self._events: Dict[KT, EventData] = {}
        self._nolock = nolock

    def __len__(self) -> int:
        total = 0
        for shard in self._shards:
            total += len(shard)
        return total

    def with_loader(self, loader: Callable[[KT], VT], ttl: Optional[timedelta]) -> None:
        self._loader = loader
        self._ttl = ttl

    def get(self, key: KT) -> Tuple[Optional[VT], bool]:
        """
        Retrieve data with cache key. If given key is not in cache, return default value.

        :param key: key hashable, use str/int for best performance.
        :param default: returned value if key is not found in cache, default None.
        """
        kh = spread(hash(key))
        idx = kh & (self._shard_count - 1)
        (v, ok) = self._shards[idx].get(key, kh)
        if ok:
            self._read_buffer.add(kh)
        return (v, ok)

    def _ttl_nano(self, ttl: Optional[timedelta]) -> int:
        ttl_ns = 0
        if ttl is not None:
            seconds = ttl.total_seconds()
            if seconds == 0:
                raise Exception("ttl must be positive")
            ttl_ns = int(seconds * 1e9)
        return ttl_ns

    def _get_or_compute(self, key: KT, fn: Callable[[], VT]) -> VT:
        kh = spread(hash(key))
        shard = self._shards[kh & (self._shard_count - 1)]
        (v, ok) = shard.get(key, kh)
        if ok:
            self._read_buffer.add(kh)
            return cast(VT, v)

        # nolock means single threading mode,
        # where cache stampe problem not exists
        if self._nolock:
            r = fn()
            ttl_ns = self._ttl_nano(self._ttl)
            shard._set_nolock(key, kh, r, ttl_ns)
            self._write_buffer.add(kh, ttl_ns)
            return r

        event = EventData(Event(), None)
        with shard._mutex:
            # check key exists again, because other threads might set the value already
            (v, ok) = shard._get_nolock(key, kh)
            if ok:
                self._read_buffer.add(kh)
                return cast(VT, v)
            ve = self._events.setdefault(key, event)
        # the shard mutex is released when computing value for the key,
        # so following get on same key can wait for the event.
        # it's safe to do so because this method is only used for decorator,
        # and user can not delete/update cache manually when using decorator
        if ve is event:
            result_sync = fn()
            event.data = result_sync
            ttl_ns = self._ttl_nano(self._ttl)
            with shard._mutex:
                shard._set_nolock(key, kh, result_sync, ttl_ns)
                self._events.pop(key, None)
            event.event.set()
            self._write_buffer.add(kh, ttl_ns)
        else:
            ve.event.wait()
        result_sync = cast(VT, ve.data)
        return result_sync

    def _get_loading(self, key: KT) -> Tuple[Optional[VT], bool]:
        if self._loader is None:
            raise Exception("loader function is None")

        return (
            self._get_or_compute(
                key, cast(Callable[[], VT], lambda: self._loader(key))
            ),
            True,
        )

    def _drain_read(self, keys: List[int]) -> None:
        try:
            self._nolock or self._core_mutex.acquire()
            self.core.access(keys)
        except Exception as e:
            raise e
        finally:
            if not self._nolock:
                self._core_mutex.release()

    def _drain_write(self, entries: List[Tuple[int, int]]) -> List[int]:
        return self.core.set(entries)

    def _remove_keys(self, keys: List[int]) -> None:
        for key in keys:
            self._shards[key & (self._shard_count - 1)].remove(key)

    # send writer buffer to policy immediately, used in test and clear
    def _force_drain_write(self) -> None:
        try:
            self._nolock or self._core_mutex.acquire()
            wb = self._write_buffer.buffer[: self._write_buffer.tail]
            if len(wb) > 0:
                evicted = self.core.set(wb)
                for key in evicted:
                    self._shards[key & (self._shard_count - 1)].remove(key)
                self._write_buffer.tail = 0
        except Exception as e:
            raise e
        finally:
            if not self._nolock:
                self._core_mutex.release()

    # used in Django adapter touch method, sets a new expiration for a key
    def _access(self, key: KT, ttl: Optional[timedelta] = None) -> None:
        kh = spread(hash(key))
        ttl_ns = self._ttl_nano(ttl)
        self._shards[kh & (self._shard_count - 1)].set_ttl(key, ttl_ns)
        self._write_buffer.add(kh, ttl_ns)

    def set(self, key: KT, value: VT, ttl: Optional[timedelta] = None) -> None:
        """
        Add new data to cache. If the key already exists, value will be overwritten.

        :param key: key hashable, use str/int for best performance.
        :param value: cached value.
        :param ttl: timedelta to store the data. Default is None which means no expiration. Value smaller than 1 second will round to 1 second. Set a negative value will panic.
        """
        kh = spread(hash(key))

        ttl_ns = self._ttl_nano(ttl)

        self._shards[kh & (self._shard_count - 1)].set(key, kh, value, ttl_ns)
        self._write_buffer.add(kh, ttl_ns)

    def delete(self, key: KT) -> bool:
        """
        Remove key from cache. Return True if given key exists in cache and been deleted.

        :param key: key hashable, use str/int for best performance.
        """
        kh = spread(hash(key))
        success = self._shards[kh & (self._shard_count - 1)].remove(kh)
        self._write_buffer.add(kh, -1)
        return success

    async def maintenance(self) -> None:
        """
        Remove expired keys.
        """
        while True:
            evicted: List[int] = []
            with self._core_mutex:
                if self._closed:
                    return
                evicted = self.core.advance()
            for key in evicted:
                self._shards[key & (self._shard_count - 1)].remove_expired(key)
            await asyncio.sleep(1)

    def clear(self) -> None:
        self._force_drain_write()
        with self._core_mutex:
            self.core.clear()
        for shard in self._shards:
            shard.clear()

    def close(self) -> None:
        with self._core_mutex:
            self._closed = True

    def __del__(self) -> None:
        self.clear()
        self.close()

    def stats(self) -> CacheStats:
        misses = 0
        hits = 0
        for shard in self._shards:
            with shard._mutex:
                hits += next(shard._hits)
                misses += next(shard._misses)
        return CacheStats(hits + misses, hits)



import datetime
import time
from typing import Iterable, List, cast

import pytest
from django.core.cache import BaseCache
from django.core.cache import cache as default_cache

from theine.adapters.django import Cache as Theine


@pytest.fixture
def cache() -> Iterable[BaseCache]:
    yield default_cache
    default_cache.clear()


class TestTheineCache:
    def test_settings(self, cache: BaseCache) -> None:
        assert cache._max_entries == 1000
        assert cache.default_timeout == 60

    def test_unicode_keys(self, cache: BaseCache) -> None:
        cache.set("ключ", "value")
        res = cache.get("ключ")
        assert res == "value"

    def test_save_and_integer(self, cache: BaseCache) -> None:
        cache.set("test_key", 2)
        res = cache.get("test_key", "Foo")

        assert isinstance(res, int)
        assert res == 2

    def test_save_string(self, cache: BaseCache) -> None:
        cache.set("test_key", "hello" * 1000)
        res = cache.get("test_key")

        assert isinstance(res, str)
        assert res == "hello" * 1000

        cache.set("test_key", "2")
        res = cache.get("test_key")

        assert isinstance(res, str)
        assert res == "2"

    def test_save_unicode(self, cache: BaseCache) -> None:
        cache.set("test_key", "heló")
        res = cache.get("test_key")

        assert isinstance(res, str)
        assert res == "heló"

    def test_save_dict(self, cache: BaseCache) -> None:
        now_dt = datetime.datetime.now()
        test_dict = {"id": 1, "date": now_dt, "name": "Foo"}

        cache.set("test_key", test_dict)
        res = cache.get("test_key")

        assert isinstance(res, dict)
        assert res["id"] == 1
        assert res["name"] == "Foo"
        assert res["date"] == now_dt

    def test_save_float(self, cache: BaseCache) -> None:
        float_val = 1.345620002

        cache.set("test_key", float_val)
        res = cache.get("test_key")

        assert isinstance(res, float)
        assert res == float_val

    def test_timeout(self, cache: BaseCache) -> None:
        cache.set("test_key", 222, timeout=3)
        time.sleep(4)

        res = cache.get("test_key")
        assert res is None

    def test_timeout_0(self, cache: BaseCache) -> None:
        cache.set("test_key", 222, timeout=0)
        res = cache.get("test_key")
        assert res is None

    def test_timeout_parameter_as_positional_argument(self, cache: BaseCache) -> None:
        cache.set("test_key", 222, -1)
        res = cache.get("test_key")
        assert res is None

        cache.set("test_key", 222, 1)
        res1 = cache.get("test_key")
        time.sleep(2)
        res2 = cache.get("test_key")
        assert res1 == 222
        assert res2 is None

    def test_timeout_negative(self, cache: BaseCache) -> None:
        cache.set("test_key", 222, timeout=-1)
        res = cache.get("test_key")
        assert res is None

        cache.set("test_key", 222, timeout=None)
        cache.set("test_key", 222, timeout=-1)
        res = cache.get("test_key")
        assert res is None

    def test_timeout_tiny(self, cache: BaseCache) -> None:
        cache.set("test_key", 222, timeout=0.00001)
        res = cache.get("test_key")
        assert res in (None, 222)

    def test_set_add(self, cache: BaseCache) -> None:
        cache.set("add_key", "Initial value")
        assert cache.add("add_key", "New value") is False

        assert cache.get("add_key") == "Initial value"
        assert cache.add("other_key", "New value") is True

    def test_get_many(self, cache: BaseCache) -> None:
        cache.set("a", 1)
        cache.set("b", 2)
        cache.set("c", 3)

        res = cache.get_many(["a", "b", "c"])
        assert res == {"a": 1, "b": 2, "c": 3}

    def test_get_many_unicode(self, cache: BaseCache) -> None:
        cache.set("a", "1")
        cache.set("b", "2")
        cache.set("c", "3")

        res = cache.get_many(["a", "b", "c"])
        assert res == {"a": "1", "b": "2", "c": "3"}

    def test_set_many(self, cache: BaseCache) -> None:
        cache.set_many({"a": 1, "b": 2, "c": 3})
        res = cache.get_many(["a", "b", "c"])
        assert res == {"a": 1, "b": 2, "c": 3}

    def test_delete(self, cache: BaseCache) -> None:
        cache.set_many({"a": 1, "b": 2, "c": 3})
        assert cache.delete("a") is True
        assert cache.get_many(["a", "b", "c"]) == {"b": 2, "c": 3}
        assert cache.delete("a") is False

    def test_delete_many(self, cache: BaseCache) -> None:
        cache.set_many({"a": 1, "b": 2, "c": 3})
        cache.delete_many(["a", "b"])
        assert cache.get_many(["a", "b", "c"]) == {"c": 3}

    def test_delete_many_generator(self, cache: BaseCache) -> None:
        cache.set_many({"a": 1, "b": 2, "c": 3})
        cache.delete_many(key for key in ["a", "b"])
        res = cache.get_many(["a", "b", "c"])
        assert res == {"c": 3}

    def test_delete_many_empty_generator(self, cache: BaseCache) -> None:
        cache.delete_many(key for key in cast(List[str], []))

    def test_incr(self, cache: BaseCache) -> None:
        cache.set("num", 1)
        cache.incr("num")
        res = cache.get("num")
        assert res == 2

        cache.incr("num", 10)
        res = cache.get("num")
        assert res == 12

        # max 64 bit signed int
        cache.set("num", 9223372036854775807)

        cache.incr("num")
        res = cache.get("num")
        assert res == 9223372036854775808

        cache.incr("num", 2)
        res = cache.get("num")
        assert res == 9223372036854775810

        cache.set("num", 3)

        cache.incr("num", 2)
        res = cache.get("num")
        assert res == 5

    def test_incr_no_timeout(self, cache: BaseCache) -> None:
        cache.set("num", 1, timeout=None)

        cache.incr("num")
        res = cache.get("num")
        assert res == 2

        cache.incr("num", 10)
        res = cache.get("num")
        assert res == 12

        # max 64 bit signed int
        cache.set("num", 9223372036854775807, timeout=None)

        cache.incr("num")
        res = cache.get("num")
        assert res == 9223372036854775808

        cache.incr("num", 2)
        res = cache.get("num")
        assert res == 9223372036854775810

        cache.set("num", 3, timeout=None)

        cache.incr("num", 2)
        res = cache.get("num")
        assert res == 5

    def test_get_set_bool(self, cache: BaseCache) -> None:
        cache.set("bool", True)
        res = cache.get("bool")

        assert isinstance(res, bool)
        assert res is True

        cache.set("bool", False)
        res = cache.get("bool")

        assert isinstance(res, bool)
        assert res is False

    def test_version(self, cache: BaseCache) -> None:
        cache.set("keytest", 2, version=2)
        res = cache.get("keytest")
        assert res is None

        res = cache.get("keytest", version=2)
        assert res == 2

    def test_incr_version(self, cache: BaseCache) -> None:
        cache.set("keytest", 2)
        cache.incr_version("keytest")

        res = cache.get("keytest")
        assert res is None

        res = cache.get("keytest", version=2)
        assert res == 2

    def test_ttl_incr_version_no_timeout(self, cache: BaseCache) -> None:
        cache.set("my_key", "hello world!", timeout=None)

        cache.incr_version("my_key")

        my_value = cache.get("my_key", version=2)

        assert my_value == "hello world!"

    def test_touch_zero_timeout(self, cache: BaseCache) -> None:
        cache.set("test_key", 222, timeout=10)

        assert cache.touch("test_key", 0) is True
        res = cache.get("test_key")
        assert res is None

    def test_touch_positive_timeout(self, cache: BaseCache) -> None:
        cache.set("test_key", 222, timeout=10)

        assert cache.touch("test_key", 2) is True
        assert cache.get("test_key") == 222
        time.sleep(3)
        assert cache.get("test_key") is None

    def test_touch_negative_timeout(self, cache: BaseCache) -> None:
        cache.set("test_key", 222, timeout=10)

        assert cache.touch("test_key", -1) is True
        res = cache.get("test_key")
        assert res is None

    def test_touch_missed_key(self, cache: BaseCache) -> None:
        assert cache.touch("test_key_does_not_exist", 1) is False

    def test_touch_forever(self, cache: Theine) -> None:
        cache.set("test_key", "foo", timeout=1)
        result = cache.touch("test_key", None)
        assert result is True
        time.sleep(2)
        assert cache.get("test_key") == "foo"

    def test_touch_forever_nonexistent(self, cache: BaseCache) -> None:
        result = cache.touch("test_key_does_not_exist", None)
        assert result is False

    def test_touch_default_timeout(self, cache: BaseCache) -> None:
        cache.set("test_key", "foo", timeout=1)
        result = cache.touch("test_key")
        assert result is True
        time.sleep(2)
        assert cache.get("test_key") == "foo"

    def test_clear(self, cache: BaseCache) -> None:
        cache.set("foo", "bar")
        value_from_cache = cache.get("foo")
        assert value_from_cache == "bar"
        cache.clear()
        value_from_cache_after_clear = cache.get("foo")
        assert value_from_cache_after_clear is None



from typing import TypeVar, Generic, Hashable

KT = TypeVar("KT", bound=Hashable)
VT = TypeVar("VT")


class CacheStats:
    def __init__(self, total: int, hit: int):
        self.request_count = total
        self.hit_count = hit
        self.miss_count = self.request_count - self.hit_count
        self.hit_rate = self.hit_count / self.request_count


class Entry(Generic[VT]):
    __slots__ = ("value", "expire")
    value: VT
    expire: int

    def __init__(self, value: VT, expire: int) -> None:
        self.value = value
        self.expire = expire



from threading import Lock
from typing import List, Callable
from random import getrandbits
from theine.utils import round_up_power_of_2
from contextlib import nullcontext
import os


# a striped lossy buffer
class StripedBuffer:

    def __init__(self, clear_buffer: Callable[[List[int]], None], nolock: bool = False):
        self.buffer_count = 4 * round_up_power_of_2(os.cpu_count() or 4)
        self.buffers: List[List[int]] = [[] for _ in range(self.buffer_count)]
        self.mutexes = [Lock() for _ in range(self.buffer_count)]
        self.buffer_size = 16
        self.clear_buffer = clear_buffer
        self.nolock = nolock

    def add(self, hash_value: int) -> None:
        index = getrandbits(32) & (self.buffer_count - 1)
        waiting: List[int] = []

        # skip if acquire lock failed, lossy1
        if self.nolock or self.mutexes[index].acquire(blocking=False):
            bs = len(self.buffers[index])
            # skip if buffer items count >= buffer size, lossy2
            if bs >= self.buffer_size:
                self.mutexes[index].release()
                return
            self.buffers[index].append(hash_value)
            if bs == self.buffer_size - 1:
                # move entries from current buffer to waiting buffer, then clear current buffer,
                # so current buffer can continue accept new entry when waiting buffer is processing
                waiting = self.buffers[index]
                self.buffers[index] = []
            if not self.nolock:
                self.mutexes[index].release()

        # clear will hold global policy mutex
        if waiting:
            self.clear_buffer(waiting)



import threading
import random
from theine.write_buffer import WriteBuffer
from threading import Lock
from time import sleep


def test_write_buffer() -> None:
    cq = []
    buffer = WriteBuffer(lambda s: cq.append(s), lambda s: s, Lock(), False)
    for i in range(40):
        buffer.add(i, 0)
    assert cq == [[(i, 0)] for i in range(40)]


def test_write_buffer_busy_lock() -> None:
    cq = []
    lock = Lock()
    buffer = WriteBuffer(lambda s: cq.append(s), lambda s: s, lock, False)
    lc = 0
    lock.acquire()
    for i in range(40):
        if lc == 6:
            lock.release()
            lc = 0
            buffer.add(i, 0)
            lock.acquire()
        else:
            buffer.add(i, 0)
            lc += 1
    lock.release()
    buffer.add(999, 999)
    assert cq == [
        [(i, 0) for i in range(7)],
        [(i + 7, 0) for i in range(7)],
        [(i + 14, 0) for i in range(7)],
        [(i + 21, 0) for i in range(7)],
        [(i + 28, 0) for i in range(7)],
        [(i + 35, 0) for i in range(5)] + [(999, 999)],
    ]


def test_write_buffer_multithreaded() -> None:
    cq = []
    buffer = WriteBuffer(lambda s: cq.append(s), lambda s: s, Lock(), False)

    def add_items(start):
        for i in range(10):
            sleep(random.randint(5, 30) / 1000)
            buffer.add(start * 10 + i, 0)

    threads = []
    for start in range(4):
        thread = threading.Thread(target=add_items, args=[start])
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join()

    m = {}
    for i in cq:
        for j in i:
            m[j[0]] = True
    assert len(m) == 40
    l = list(m.keys())
    l.sort()
    assert l == list(range(40))


def test_write_buffer_busy_lock_multithreaded() -> None:
    cq = []
    lock = Lock()
    buffer = WriteBuffer(lambda s: cq.append(s), lambda s: s, lock, False)

    def add_items(start):
        for i in range(10):
            sleep(random.randint(5, 30) / 1000)
            buffer.add(start * 10 + i, 0)

    done = False

    def locking():
        nonlocal done
        while not done:
            lock.acquire()
            sleep(5 / 1000)
            lock.release()
            sleep(2 / 1000)

    ls = threading.Thread(target=locking, args=[])
    ls.start()
    threads = []
    for start in range(4):
        thread = threading.Thread(target=add_items, args=[start])
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join()
    done = True
    ls.join()
    buffer.add(999, 999)

    m = {}
    for i in cq:
        for j in i:
            m[j[0]] = True
    assert len(m) == 41
    l = list(m.keys())
    l.sort()
    assert l == list(range(40)) + [999]



import random
import time
from threading import Lock, Thread

import matplotlib.pyplot as plt
import numpy as np
from bounded_zipf import Zipf
from cachetools import TTLCache, cached
from scipy.interpolate import PchipInterpolator

from theine import Memoize

CPU = 6
DATA_LEN = 2 << 15


def key_gen():
    z = Zipf(1.0001, 9, DATA_LEN)
    keys = []
    i = 0
    while i < DATA_LEN:
        keys.append(z.get())
        i += 1

    return keys


def bench_run(index, fn, tp, reporter=None, keys=[]):
    if len(keys) == 0:
        keys = key_gen()
    else:
        keys = keys[:]

    start = random.randint(0, DATA_LEN - 1)
    s = time.monotonic_ns()

    for i in range(DATA_LEN):
        if tp == 0:
            fn(keys[(i + start) & DATA_LEN - 1])
        else:
            fn(index * DATA_LEN + i)

    dt = time.monotonic_ns() - s
    if reporter is None:
        nop = dt / DATA_LEN
        print(f"single thread {'read' if tp==0 else 'write'}: {nop:.2f} ns/op")
    else:
        reporter(DATA_LEN)


def bench_run_parallel(r, w, tp):
    if tp == "theine":
            @Memoize(DATA_LEN * 2, None)
            def get(key):
                return key

            @get.key
            def k(key):
                return str(key)
    elif tp == "cachetools":
        @cached(cache=TTLCache(maxsize=DATA_LEN * 2, ttl=20000), lock=Lock())
        def get(key):
            return key

    keys = key_gen()
    for k in keys:
        get(k)
        get(k)
        get(k)

    tl = []

    count_sum = 0
    mu = Lock()

    def report(count):
        nonlocal count_sum

        with mu:
            count_sum += count

    s = time.monotonic_ns()
    for i in range(r + w):
        t = Thread(target=bench_run, args=[i, get, 0 if i < r else 1, report, keys])
        tl.append(t)
        t.start()

    for t in tl:
        t.join()
    dt = time.monotonic_ns() - s

    nop = dt / count_sum
    print(f"read-{r} write-{w}: {nop:.2f} ns/op")
    return 1 / nop * 1e9


def bench_run_nolock(tp):
    if tp == "theine":
            @Memoize(DATA_LEN * 2, None, nolock=True)
            def get(key):
                return key

            @get.key
            def k(key):
                return str(key)
    elif tp == "cachetools":
        @cached(cache=TTLCache(maxsize=DATA_LEN * 2, ttl=20000), info=True)
        def get(key):
            return key

    keys = key_gen()
    for k in keys:
        get(k)
        get(k)
        get(k)

    count_sum = 0
    mu = Lock()

    def report(count):
        nonlocal count_sum

        with mu:
            count_sum += count

    s = time.monotonic_ns()
    bench_run(0, get, 0, report, keys)
    dt = time.monotonic_ns() - s
    nop = dt / count_sum
    ops = 1 / nop * 1e9
    print(f"{tp} 100% read: {nop} ns/op {ops} ops/s")
    return 1 / nop * 1e9

def init_plot(title):
    fig, ax = plt.subplots()
    ax.set_xlabel("threads")
    ax.set_ylabel("ops/s")
    ax.set_title(title)
    fig.set_figwidth(12)
    return fig, ax

def bench_and_plot(parallel, name):
    x = []
    y_theine = []
    y_cachetools = []

    for p in parallel:
        x.append(p)
        y_theine.append(bench_run_parallel(p, 0, "theine"))
        y_cachetools.append(bench_run_parallel(p, 0, "cachetools"))

    fig, ax = init_plot("Read Throughput")
    # Smooth the lines using np.interp for interpolation
    x_new = np.linspace(min(x), max(x), 300)

    theine_spl = PchipInterpolator(x, y_theine)
    cachetools_spl = PchipInterpolator(x, y_cachetools)
    y_theine_smooth = theine_spl(x_new)
    y_cachetools_smooth = cachetools_spl(x_new)

    ax.plot(x_new, y_theine_smooth, "b-", label="theine")
    ax.plot(x_new, y_cachetools_smooth, "r-", label="cachetools")
    ax.plot(x, y_theine, "bo")
    ax.plot(x, y_cachetools, "ro")

    ax.legend()
    fig.savefig(f"benchmarks/{name.lower()}.png", dpi=200)


def bench_and_plot_nolock(name):
    bench_run_nolock("theine")
    bench_run_nolock("cachetools")

if __name__ == "__main__":
    bench_and_plot_nolock("read_throughput")
    bench_and_plot([1,2,4,8,12,16,20,26,32], "read_throughput")



import sysconfig
import sys
import asyncio
import concurrent.futures
from datetime import timedelta
from random import randint
from threading import Thread
from time import sleep
from typing import Any, Dict, List
from unittest.mock import Mock

import pytest
from bounded_zipf import Zipf  # type: ignore[import]

from theine.theine import Memoize

EXCEPTION_THROWING_CACHE_TTL = timedelta(seconds=3)

is_freethreaded = bool(sysconfig.get_config_var("Py_GIL_DISABLED"))


@Memoize(1000, None)
def foo(id: int, m: Mock) -> Dict[str, int]:
    m(id)
    return {"id": id}


@foo.key
def _(id: int, m: Mock) -> str:
    return f"id-{id}"


@Memoize(1000, None)
def foo_empty() -> Dict:
    return {"id": "empty"}


@foo_empty.key
def _() -> str:
    return "empty"


@Memoize(1000, None)
async def async_foo(id: int, m: Mock) -> Dict:
    m(id)
    await asyncio.sleep(1)
    return {"id": id}


@async_foo.key
def _(id: int, m: Mock) -> str:
    return f"id-{id}"


@Memoize(1000, ttl=EXCEPTION_THROWING_CACHE_TTL)
async def async_foo_that_can_raise(input_val: int, m: Mock) -> Dict:
    m(input_val)
    await asyncio.sleep(1)
    return {"id": input_val}


@async_foo_that_can_raise.key
def _(input_val: int, m: Mock) -> str:
    return f"id-{input_val}"


class Bar:
    @Memoize(1000, None)
    def foo(self, id: int, m: Mock) -> Dict:
        m(id)
        return {"id": id}

    @foo.key
    def _(self, id: int, m: Mock) -> str:
        return f"id-{id}"

    @Memoize(1000, None)
    async def async_foo(self, id: int, m: Mock) -> Dict:
        m(id)
        await asyncio.sleep(1)
        return {"id": id}

    @async_foo.key
    def _(self, id: int, m: Mock) -> str:
        return f"id-{id}"

    @Memoize(1000, None)
    def foo_empty(self) -> str:
        return "empty"

    @foo_empty.key
    def _(self) -> str:
        return "empty"

    @Memoize(1000, None)
    @classmethod
    def foo_class(cls, id: int, m: Mock) -> Dict:
        m(id)
        return {"id": id}

    @foo_class.key
    def _(cls: Any, id: int, m: Mock) -> str:
        m(id)
        return f"id-{id}"

    @Memoize(1000, None)
    def foo_auto(self, id: int, m: Mock) -> Dict:
        m(id)
        return {"id": id}

    @Memoize(1000, None)
    async def async_foo_auto(self, id: int, m: Mock) -> Dict:
        m(id)
        await asyncio.sleep(1)
        return {"id": id}


def test_sync_decorator() -> None:
    mock = Mock()
    threads: List[Thread] = []
    assert foo.__name__ == "foo"  # type: ignore

    def assert_id(id: int, m: Mock):
        try:
            v = foo(id, m)
        except Exception as e:
            print(e)
        assert v["id"] == id

    for _ in range(500):
        t = Thread(target=assert_id, args=[randint(0, 5), mock])
        threads.append(t)
        t.start()

    for t in threads:
        t.join()

    assert mock.call_count == 6
    ints = [i[0][0] for i in mock.call_args_list]
    assert set(ints) == {0, 1, 2, 3, 4, 5}


def test_sync_decorator_empty() -> None:
    threads: List[Thread] = []

    def assert_id():
        assert foo_empty()["id"] == "empty"

    for _ in range(500):
        t = Thread(target=assert_id, args=[])
        threads.append(t)
        t.start()

    for t in threads:
        t.join()


@pytest.mark.asyncio
async def test_async_decorator() -> None:
    mock = Mock()
    assert async_foo.__name__ == "async_foo"  # type: ignore

    async def assert_id(id: int, m: Mock):
        data = await async_foo(id, m)
        assert data["id"] == id

    await asyncio.gather(*[assert_id(randint(0, 5), mock) for _ in range(500)])

    assert mock.call_count == 6
    ints = [i[0][0] for i in mock.call_args_list]
    assert set(ints) == {0, 1, 2, 3, 4, 5}


@pytest.mark.asyncio
async def test_async_decorator_with_exceptions() -> None:
    exception_to_raise = Exception("Test exception")
    assert async_foo_that_can_raise.__name__ == "async_foo_that_can_raise"  # type: ignore

    async def assert_id(id: int, m: Mock):
        data = await async_foo_that_can_raise(id, m)
        assert data["id"] == id

    # verify exception is properly raised on first call
    mock = Mock(side_effect=exception_to_raise)
    with pytest.raises(Exception) as exc_info:
        await async_foo_that_can_raise(1, mock)
    assert exc_info.value == exception_to_raise
    assert mock.call_count == 1

    # prior call should not be cached due to exception and this will run
    mock = Mock(return_value=None)
    result = await async_foo_that_can_raise(1, mock)
    assert result["id"] == 1
    assert mock.call_count == 1

    # good value should be cached and won't hit service
    mock = Mock(return_value=None)
    result = await async_foo_that_can_raise(1, mock)
    assert result["id"] == 1
    assert mock.call_count == 0

    # sleep enough to let cache expire and verify new exceptions are raised again
    await asyncio.sleep(EXCEPTION_THROWING_CACHE_TTL.total_seconds() + 0.1)
    mock = Mock(side_effect=exception_to_raise)
    with pytest.raises(Exception) as exc_info:
        await async_foo_that_can_raise(1, mock)
    assert exc_info.value == exception_to_raise
    assert mock.call_count == 1


@pytest.mark.asyncio
async def test_async_decorator_same_exception_herd() -> None:
    total = 500
    mock = Mock()

    def throwing_side_effect(*args, **kwargs):
        raise Exception("Test exception")

    mock.side_effect = throwing_side_effect

    coros = [async_foo(20, mock) for _ in range(total)]
    results = await asyncio.gather(*coros, return_exceptions=True)

    # should all be exceptions
    assert all([isinstance(res, Exception) for res in results])

    # should all be the same exception
    uniq_exceptions = set(results)
    assert len(uniq_exceptions) == 1

    # herd should have only result in 1 call
    assert mock.call_count == 1

    # verify if these get awaited on again for some reason same exception is raised
    for coro in coros:
        with pytest.raises(Exception) as exc_info:
            await coro
        assert exc_info.value in uniq_exceptions


def test_instance_method_sync() -> None:
    mock = Mock()
    threads: List[Thread] = []
    bar = Bar()
    assert bar.foo.__name__ == "foo"  # type: ignore

    def assert_id(id: int, m: Mock):
        assert bar.foo(id, m)["id"] == id

    for _ in range(500):
        t = Thread(target=assert_id, args=[randint(0, 5), mock])
        threads.append(t)
        t.start()

    for t in threads:
        t.join()

    assert mock.call_count == 6
    ints = [i[0][0] for i in mock.call_args_list]
    assert set(ints) == {0, 1, 2, 3, 4, 5}


@pytest.mark.asyncio
async def test_instance_method_async() -> None:
    mock = Mock()
    bar = Bar()
    assert bar.async_foo.__name__ == "async_foo"  # type: ignore

    async def assert_id(id: int, m: Mock):
        data = await bar.async_foo(id, m)
        assert data["id"] == id

    await asyncio.gather(*[assert_id(randint(0, 5), mock) for _ in range(500)])

    assert mock.call_count == 6
    ints = [i[0][0] for i in mock.call_args_list]
    assert set(ints) == {0, 1, 2, 3, 4, 5}


@Memoize(1000, None)
def foo_auto_key(a: int, b: int, c: int = 5) -> Dict:
    return {"a": a, "b": b, "c": c}


def test_auto_key() -> None:
    tests = [
        ([1, 2, 3], {}, (1, 2, 3)),
        ([1, 2], {}, (1, 2, 5)),
        ([1], {"b": 2}, (1, 2, 5)),
        ([], {"a": 1, "b": 2}, (1, 2, 5)),
        ([], {"a": 1, "b": 2, "c": 3}, (1, 2, 3)),
    ]

    def assert_data(args, kwargs, expected):
        result = foo_auto_key(*args, **kwargs)
        assert result["a"] == expected[0]
        assert result["b"] == expected[1]
        assert result["c"] == expected[2]

    for case in tests:
        assert_data(*case)


@Memoize(1000, None)
async def async_foo_auto_key(a: int, b: int, c: int = 5) -> Dict:
    return {"a": a, "b": b, "c": c}


@pytest.mark.asyncio
async def test_auto_key_async() -> None:
    tests = [
        ([1, 2, 3], {}, (1, 2, 3)),
        ([1, 2], {}, (1, 2, 5)),
        ([1], {"b": 2}, (1, 2, 5)),
        ([], {"a": 1, "b": 2}, (1, 2, 5)),
        ([], {"a": 1, "b": 2, "c": 3}, (1, 2, 3)),
    ]

    async def assert_data(args, kwargs, expected):
        result = await async_foo_auto_key(*args, **kwargs)
        assert result["a"] == expected[0]
        assert result["b"] == expected[1]
        assert result["c"] == expected[2]

    for case in tests:
        await assert_data(*case)


def test_instance_method_auto_key_sync() -> None:
    mock = Mock()
    threads: List[Thread] = []
    bar = Bar()

    def assert_id(id: int, m: Mock):
        assert bar.foo_auto(id, m)["id"] == id

    for _ in range(500):
        t = Thread(target=assert_id, args=[randint(0, 5), mock])
        threads.append(t)
        t.start()

    for t in threads:
        t.join()

    assert mock.call_count == 6
    ints = [i[0][0] for i in mock.call_args_list]
    assert set(ints) == {0, 1, 2, 3, 4, 5}


@pytest.mark.asyncio
async def test_instance_method_auto_key_async() -> None:
    mock = Mock()
    bar = Bar()

    async def assert_id(id: int, m: Mock):
        data = await bar.async_foo_auto(id, m)
        assert data["id"] == id

    await asyncio.gather(*[assert_id(randint(0, 5), mock) for _ in range(500)])

    assert mock.call_count == 6
    ints = [i[0][0] for i in mock.call_args_list]
    assert set(ints) == {0, 1, 2, 3, 4, 5}


@Memoize(1000, timedelta(seconds=1))
def foo_to(id: int) -> Dict:
    return {"id": id}


@foo_to.key
def _(id: int) -> str:
    return f"id-{id}"


def test_ttl() -> None:
    for i in range(30):
        result = foo_to(i)
        assert result["id"] == i
    assert len(foo_to._cache) == 30
    foo_to._cache._force_drain_write()
    current = 30
    counter = 0
    while True:
        sleep(5)
        counter += 1
        assert len(foo_to._cache) < current
        current = len(foo_to._cache)
        if current == 0:
            break


@Memoize(1000, timedelta(seconds=1))
def foo_to_auto(id: int, m: Mock) -> Dict:
    m(id)
    return {"id": id}


def test_ttl_auto_key() -> None:
    mock = Mock()
    for i in range(30):
        result = foo_to_auto(i, mock)
        assert result["id"] == i
    assert len(foo_to_auto._cache) == 30
    foo_to_auto._cache._force_drain_write()
    current = 30
    counter = 0
    while True:
        sleep(5)
        counter += 1
        assert len(foo_to._cache) < current
        current = len(foo_to._cache)
        if current == 0:
            break
    assert len(foo_to_auto._cache.core.keys()) == 0


def test_cache_full_evict() -> None:
    mock = Mock()
    for i in range(30, 1500):
        result = foo_to_auto(i, mock)
        assert result["id"] == i
    foo_to_auto._cache._force_drain_write()
    assert len(foo_to_auto._cache) == 1000


def test_cache_full_auto_key_sync_multi() -> None:
    mock = Mock()
    threads: List[Thread] = []

    def assert_id(id: int, m: Mock):
        assert foo_to_auto(id, m)["id"] == id

    for i in range(2000, 3500):
        t = Thread(target=assert_id, args=[i, mock])
        threads.append(t)
        t.start()

    for t in threads:
        t.join()

    foo_to_auto._cache._force_drain_write()
    assert len(foo_to_auto._cache) == 1000


@Memoize(1000, ttl=None)
def read_auto_key(key: str) -> str:
    return key


def assert_read_key(n: int) -> None:
    key = f"key:{n}"
    v = read_auto_key(key)
    assert v == key
    assert len(read_auto_key._cache) < 2000
    print(".", end="")


def zipf_key_gen(total):
    z = Zipf(1.01, 9.0, 50000 * 1000)
    for _ in range(total):
        yield z.get()

def test_cocurrency_load() -> None:
    z = Zipf(1.0001, 10, 5000000)
    with concurrent.futures.ThreadPoolExecutor(max_workers=1000) as executor:
        for _ in range(200000):
            future = executor.submit(assert_read_key, z.get())
            exception = future.exception()
            if exception:
                raise exception
    print("==== done ====", len(read_auto_key._cache))

@pytest.fixture(params=[True, False])
def nolock(request):
    return request.param

def test_sync_decorator_zipf(nolock) -> None:
    miss = 0

    @Memoize(50000, ttl=None, nolock=nolock)
    def read(key: int) -> int:
        nonlocal miss
        miss += 1
        return key

    for key in zipf_key_gen(2000000):
        v = read(key)
        assert key == v

    cache = read._cache
    cache._force_drain_write()
    stats = cache.stats()
    assert stats.request_count == 2000000
    assert stats.hit_rate > 0.5 and stats.hit_rate < 0.6
    assert 1 - (miss / 2000000) > 0.5 and 1 - (miss / 2000000) < 0.6

def test_sync_decorator_zipf_correctness(nolock) -> None:
    for size in [500, 2000, 10000, 50000]:
        @Memoize(size, ttl=None, nolock=nolock)
        def read(key: int) -> int:
            return key

        for key in zipf_key_gen(1000000):
            v = read(key)
            assert key == v

        cache = read._cache
        cache._force_drain_write()
        total = 0
        for shard in cache._shards:
            total += len(shard._map)
        info = cache.core.debug_info()
        assert total == info.len
        assert info.window_len + info.probation_len + info.protected_len == total
        keys = cache.core.keys()
        assert len(keys) == total
        for kh in cache.core.keys():
            idx = kh & (cache._shard_count - 1)
            assert kh in cache._shards[idx]._key_map

def test_sync_decorator_zipf_correctness_parallel() -> None:
    def run(read):
        if is_freethreaded:
            assert sys._is_gil_enabled() == False

        keys = list(zipf_key_gen(500000))
        for key in keys:
            v = read(key)
            assert key == v

    for size in [500, 2000, 10000, 50000]:
        @Memoize(size, ttl=None)
        def read(key: int) -> int:
            return key

        threads = []
        for start in range(4):
            thread = Thread(target=run, args=[read])
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        cache = read._cache
        cache._force_drain_write()
        total = 0
        for shard in cache._shards:
            total += len(shard._map)
        info = cache.core.debug_info()
        assert total == info.len
        assert info.window_len + info.probation_len + info.protected_len == total
        keys = cache.core.keys()
        assert len(keys) == total
        for kh in cache.core.keys():
            idx = kh & (cache._shard_count - 1)
            assert kh in cache._shards[idx]._key_map
